*-- File: ARRAYEX.CC

CLASS ArrayEX 
*----------------------------------------------------------------------
*-- Description:
*
*      This class contains a set of array routines, and is designed 
*      to enhance the standard array class in Visual dBASE. This is
*      not a class derived from array!
*
*-- Programmers:
*
*   Ken Mayer
*   Adam L. Menkes
*   Jay Parsons (CIS: 72662,1302)
*   Werner Borsbach (CIS:100010,2236)
*   Bowen Moursund (CIS: 72662,436)
*   Marcello Perathoner (CIS: 100752,3200)
*
*-- History: 
*
*   11/27/1996 -- original version of these routines as a class.
*   01/11/1997 -- Updated/fixed by Marcello to handle consistency
*                 and bugs ...
*
*-- Instantiation:
*
*   set procedure to arrayex.cc additive
*   oMyArray = new ArrayEx()
*
*   Typical usage would be to call individual methods from here,
*   such as:
*
*     ? oMyArray.AAppend( cFileName, aArrayName )
*
*-- Methods:
*     Note, more details are in the headers for each method, including
*     a listing of parameters, and what they mean, as well as the
*     complete history (as we know it) for each method.
*
*--    Init()   
*        handles any "initialization" necessary, is handled
*        when a new instance of the Convert class is created.
*
*--    Release() 
*        releases this class and other classes if they weren't
*        instantiated previously
*
*--    ADiff()
*        Compares two arrays, returns match/mismatch information
*
*--    ADiffA()
*        Compares two arrays, returns match/mismatch array
*
*--    Avg()
*        Averages row or column of an array.
*
*--    Copy()
*        Copies an array, returns pointer to new array "aCopy"
*
*--    Flip()
*        Rotates array so rows become columns and vice versa
*
*--    Rec2Arr()
*        Creates an array object initialized to record of current dbf
*        (blank or filled)
*
*--    Shuffle()
*        Shuffles elements of a one-dimensional array
*
*--    Txt2Arr()
*        Adds text file to array
*
*--    Variance()
*        Variance of a row or column of array
*
*-- Dependencies:
* 
*     MATHEX.CC
*     DUFLP.H
*----------------------------------------------------------------------

   class::Init()

   FUNCTION Init
   *-------------------------------------------------------------------
   *-- Programmer..: Ken Mayer
   *-- Date........: 11/10/1996
   *-- Notes.......: Ensures that routines needed are available
   *--               It is required that this be used (see instructions
   *--               above) for this class to function properly.
   *-- Written for.: Visual dBASE 5.5
   *-- Rev. History: 11/10/1996 -- Original Code
   *-- Calls.......: math.cc
   *-- Usage.......: arrayex.Init()
   *-- Example.....: (Called on instantiation of object)
   *-- Returns.....: Nul
   *-- Parameters..: None
   *-------------------------------------------------------------------
   
      if .not. SetProc( "MATHEX.CC" )
         set procedure to mathex.cc additive
         this.SetMath = .t.
      else
         this.SetMath = .f.
      endif
      this.Math        = new MathEx()
      this.Math.parent = this
      
   RETURN NULL
   *-- EoM: Init()
   
   FUNCTION Release
   *-------------------------------------------------------------------
   *-- Programmer..: Ken Mayer
   *-- Date........: 11/27/1996
   *-- Notes.......: Used to release this object from memory.
   *-- Written for.: Visual dBASE 5.5
   *-- Rev. History: 11/24/1996 -- Original Code
   *-- Calls.......: none
   *-- Usage.......: array.Release()
   *-- Example.....: oMyArray.Release()
   *--               oMyArray = NULL  && not required, good idea
   *-- Returns.....: Null
   *-- Parameters..: None
   *-------------------------------------------------------------------

      if this.SetMath
         close procedure mathex.cc
      endif
      release object this
   
   RETURN NULL
   *-- EoM: Release()

   FUNCTION ADiff ( oArray1, oArray2, nLen )
   *-------------------------------------------------------------------
   *-- Programmer..: Werner Borsbach (CIS:100010,2236)
   *-- Date........: 11/10/1996
   *-- Notes.......: Compares two arrays, returns index of first 
   *--               differing element. 0 if all match.
   *--               Ignores all elements of either array past <nLen>.
   *-- Written for.: dBASE IV v1.5
   *-- Rev. History: 02/09/1993 Original Release as CmpArray procedure
   *--               05/16/1993 modified for use with - modified - 
   *--                          aRec2Arr()
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302                                     ,
   *--               11/10/1996 -- Modified to work with this class.
   *-- Calls.......: None
   *-- Usage.......: arrayex.ADiff (<Array1>,<Array2>,[,<nLen>])
   *-- Example.....: a1 = aex.Rec2Arr ()
   *--               edit recno ()
   *--               a2 = aex.Rec2Arr ()
   *--               ? aex.ADiff (a1, a2, a1.size)
   *-- Returns.....: 0 if arrays match, else index of first differing 
   *--               element
   *-- Parameters..: a1, a2 = arrays to compare 
   *--               nLen = Number of elements
   *-------------------------------------------------------------------
      
      local i
      private a1, a2	&& need private for type ()
      a1 = oArray1
      a2 = oArray2
      
      IF PCount () = 3
         nLen = min (nLen, a1.size)
         nLen = min (nLen, a2.size)
      ELSE
         nLen = min (a1.size, a2.size)
      ENDIF
      
      FOR i = 1 TO nLen
         IF Type ("a1 [i]") <> Type ("a2 [i]")
            RETURN i
         ENDIF
         IF a1 [i] <> a2 [i]
            RETURN i
         ENDIF
      NEXT
      
   RETURN 0
   *-- EoM: ADiff()
   
   FUNCTION ADiffA ( oArray1, oArray2, nLen )
   *-------------------------------------------------------------------
   *-- Programmer..: Werner Borsbach (CIS:100010,2236)
   *-- Date........: 11/10/1996
   *-- Notes.......: Compares two arrays, returns an array 
   *--               containing .t. for every element in that  
   *--               the arrays match.
   *--               Ignores all elements of either array past <nLen>.
   *-- Written for.: dBASE IV v1.5
   *-- Rev. History: 02/09/1993 Original Release as CmpArray procedure
   *--               05/16/1993 modified for use with - modified - 
   *--                          aRec2Arr()
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302                                     ,
   *--               11/10/1996 -- Modified to work with this class.
   *-- Calls.......: None
   *-- Usage.......: arrayex.ADiff (<Array1>,<Array2>[,<nLen>])
   *-- Example.....: a1 = aex.Rec2Arr ()
   *--               edit recno ()
   *--               a2 = aex.Rec2Arr ()
   *--               a3 = aex.ADiff (a1, a2, a1.size)
   *-- Returns.....: array 
   *-- Parameters..: a1, a2 = arrays to compare 
   *--               nLen = Number of elements
   *-------------------------------------------------------------------
      
      local i
      private a1, a2	&& need private for type ()
      a1 = oArray1
      a2 = oArray2
      
      IF PCount () = 3
         nLen = min (nLen, a1.size)
         nLen = min (nLen, a2.size)
      ELSE
         nLen = min (a1.size, a2.size)
      ENDIF

      aDiff = new array (nLen)
      
      FOR i = 1 TO nLen
         aDiff [i] = (Type ("a1 [i]") = Type ("a2 [i]")) ;
            .and. (a1 [i] = a2 [i])
      NEXT
      
   RETURN aDiff
   *-- EoM: ADiffA ()

   FUNCTION Avg( aName, nRowCol, nWhich )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (CIS: 72662,1302)
   *-- Date........: 11/10/1996
   *-- Notes.......: Mean of non-blank numeric or date values in 
   *--               specified row or column of a specified array. If 
   *--               the first value is a date, averages only dates.  
   *--               If first value is numeric or float, averages only 
   *--               numerics and floats.  Returns .F. if any value in 
   *--               row or column is any other type, if specified row
   *--               or column does not exist or if there are no 
   *--               nonblank values.
   *-- Written for.: dBASE 5.0 for Windows
   *-- Rev. History: 10/01/1994 -- Original Release
   *--               11/10/1996 -- Modified to work with this class.   
   *-- Calls.......: None
   *-- Usage.......: arrayex.Avg( <aName> [, <nRowCol> [, <nWhich>]] )
   *-- Example.....: ? oMyArray.Avg( Myarray, 1, 1 )
   *-- Returns.....: A numeric, float or date value, the mean or 
   *--               average, or .F. If any of the averaged items are 
   *--               floats, the result will be.
   *-- Parameters..: aName   = the name of the array, not in quotes.
   *--               nRowCol = the number of the row or column to 
   *--                         average, default being 1
   *--               nWhich  = one of these values:
   *--                            1 = nRowCol specifies a row (default)
   *--                            2 = nRowCol specifies a column
   *-------------------------------------------------------------------
      
      local nW, nR, xReturn, cType, nAt, nTot, nCount, xNext, cType2
      
      nW = iif( nWhich = 2, 2, 1 )
      nR = iif( pcount() > 1, nRowCol, 1 )
      xReturn = .F.
      nDims = ( &aName..dimensions )
      if nDims = 1
         if nR # 1 .or. nW # 1
            RETURN xReturn
         else
            cType = type( "aName[ nR ]" )
         endif
      else
         if nW = 2
            cType = type( "aName[1,nR]" )
         else
            cType = type( "aName[nR,1]" )
         endif
      endif
      if .not. cType $ "DFN"
         RETURN xReturn
      endif
      store 0 to nTot, nCount
      for nAt = 1 to alen( aName, nW )
          do case
             case nDims = 1
                  xNext = aName[ nAt ]
             case nW = 1
                  xNext = aName[ nR, nAt ]
             otherwise
                  xNext = aName[ nAt, nR ]
          endcase
          cType2 = type( xNext )
          if ( .not. cType2 $ "DFN" ) .or. ;
             ( ( cType = "D" ) # ( cType2 = "D" ) )
             exit
          endif
          if isblank( xNext )
             loop
          endif
          xReturn = .T.
          if cType2 = "D"
             xNext = xNext - {01/01/01}
          endif
          nTot = nTot + xNext
          nCount = nCount + 1
      endfor
      if xReturn # .F.
         xReturn = nTot/nCount + iif( cType ="D",{01/01/01},0)
      endif
      
   RETURN xReturn
   *-- EoM: Avg()

   FUNCTION Copy (oArray)
   *-------------------------------------------------------------------
   *-- Programmer..: Marcello Perathoner
   *-- Date........: 01/11/1997
   *-- Notes.......: Returns a copy of the array if 1 or 2-dimensional
   *-- Written for.: dBASE 5.0 for Windows
   *-- Rev. History: 01/11/1997 -- Original
   *-- Usage.......: arrayex.copy (oArray)
   *-- Example.....: oCopyOfArray = oArray.copy (oOriginalArray)
   *-- Returns.....: array
   *------------------------------------------------------------------

      local aCopy

      DO CASE 
         CASE oArray.dimensions = 1
            aCopy = new ARRAY (ALEN (oArray, 1))
         CASE oArray.dimensions = 2
            aCopy = new ARRAY (ALEN (oArray, 1), ALEN (oArray, 2))
         OTHERWISE
            RETURN NULL
      ENDCASE
      ACOPY (oArray, aCopy)
      
   RETURN aCopy
   *-- EoM: Copy()

   FUNCTION Flip( aName )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (CIS: 72662,1302)
   *-- Date........: 11/10/1996
   *-- Notes.......: Flips a two-dimensional array so its rows become 
   *--               its columns and vice versa, so you may then use a 
   *--               function such as array method sort() which works 
   *--               only on rows. Returns .T. if array flipped, else 
   *--               .F.
   *-- Written for.: dBASE 5.0 for Windows
   *-- Rev. History: 10/02/1994 -- original function.
   *--               11/10/1996 -- Modified to work with this class.   
   *-- Calls.......: None
   *-- Called by...: Any
   *-- Usage.......: arrayex.Flip( <aName> )
   *-- Example.....: oMyArray.Flip( aMyArray )
   *-- Returns.....: .T. if array successfully flipped, else .F.
   *-- Parameters..: aName = array name, not in quotes
   *-------------------------------------------------------------------
      
      local lSuccess, nRows, nCols, nMax, nDiag, nAt, xTemp, n1, n2
      
      lSuccess = .f.
      nCols = alen( aName, 2 )
      nRows = alen( aName, 1 )
      if nCols = 0
         RETURN lSuccess
      endif
      nMax = max( nRows, nCols )
      if nRows # nCols
         if aresize( aName, nMax, nMax, 1 ) # nMax * nMax
            RETURN lSuccess
         endif
      endif
      for nDiag = 1 to min( nRows, nCols )
          for nAt = nDiag + 1 to nMax
              xTemp = aName[ nDiag, nAt ]
              aName[ nDiag, nAt ] = aName[ nAt, nDiag ]
              aName[ nAt, nDiag ] = xTemp
          endfor
      endfor
      lSuccess = .T.
      if nRows # nCols
         lSuccess = ( aresize( aName, nCols, nRows, 1 ) = ;
                   nRows * nCols )
      endif
      
   RETURN lSuccess
   *-- EoM: Flip()

   FUNCTION Rec2Arr( lBlank )
   *-------------------------------------------------------------------
   *-- Programmer..: Bowen Moursund (CIS: 72662,436)
   *-- Date........: 11/10/1996
   *-- Notes.......: Loads this array with the field values of the
   *--               current record, or corresponding blank values. 
   *--               Memo fields are not copied to the array.
   *-- Written for.: Visual dBase 5.5
   *-- Rev. History: 05/01/1992 -- Original
   *--               05/01/1993 modified by Werner Borsbach 
   *--               (CIS: 100010,2236) - name of array is to be chosen
   *--               when the function is called so that one now can 
   *--               create multiple arrays and compare them.
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/10/1996 -- Modified to work with this class.
   *--               01/11/1997 -- Modified to leave the record ptr 
   *--                             alone and to use this array
   *-- Calls.......: None
   *-- Usage.......: arrayex.Rec2Arr(<lBlank>)
   *-- Example.....: lSuccess = oMyArray.Rec2Arr (.t.)
   *-- Returns.....: array if succesful, NULL
   *-- Parameters..: lBlank  = whether or not to create an empty array.
   *--                        .T. = blank values
   *--                        .F. = use current record values
   *-------------------------------------------------------------------
      
      local nFields, i, j, cType
      
      IF empty (dbf ())
         RETURN NULL
      ENDIF

      nFields = fldcount ()
      oArray = new array (nFields)
      IF lBlank
         j = 0 && used to skip memos, ole, binary
         FOR i = 1 TO nFields
            cType = type (field (i))
            DO CASE
               case cType = "C"
                  j = j + 1
                  oArray [j] = space (len (field (i)))
               case cType = "D"
                  j = j + 1
                  oArray [j] = {}
               case cType = "L"
                  j = j + 1
                  oArray [j] = .f.
               case cType = "N"
                  j = j + 1
                  oArray [j] = 0
               case cType = "F"
                  j = j + 1
                  oArray [j] = 0.0
            ENDCASE
         NEXT
      ELSE
         COPY TO ARRAY oArray NEXT 1
      ENDIF
       
   RETURN oArray
   *-- EoM: Rec2Arr()
   
   FUNCTION Shuffle( aName )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (CIS: 72662,1302)
   *-- Date........: 11/10/1996
   *-- Notes.......: Random shuffle of elements of a one-dimensioned 
   *--               array
   *-- Written for.: dBASE 5.0 for Windows
   *-- Rev. History: 10/01/1994 -- Original Release
   *--               11/10/1996 -- Modified to work with this class.   
   *-- Calls.......: mathex.dice()
   *-- Usage.......: arrayex.Shuffle( aName  )
   *-- Example.....: lX = oMyArray.Shuffle( MyDeck )
   *-- Returns.....: NULL
   *-- Parameters..: aName = Name of array, not in quotes
   *-- Side effects: Rearranges elements of the array
   *--               Uses some random numbers; does not reseed 
   *--               generator
   *-------------------------------------------------------------------
      
      local nLen, nAt, nRand, nTemp

      nLen = alen( aName )
      for nAt = 1 to nLen - 1
          nRand = this.Math.Dice( nLen - nAt ) + nAt
          nTemp = aName[ nRand ]
          aName[ nRand ] = aName[ nAt ]
          aName[ nAt ] = nTemp
      endfor
      
   RETURN NULL
   *-- EoM: Shuffle()

   FUNCTION Txt2Arr ( cFileName ) && FIX: changed name
   *-------------------------------------------------------------------
   *-- Programmer..: Adam L. Menkes
   *-- Date........: 11/10/1996
   *-- Notes.......: Stuffs each line of a text file into one array 
   *--               cell. File must be an ASCII Txt formatted file.
   *--               Taken from Technotes, April, 1992.
   *-- Written for.: dBASE 5.0 for Windows
   *-- Rev. History: 04/01/1992 -- Original
   *--               10/26/1993 -- Angus Scott-Fleming, release 
   *--                             "ALL LIKE"
   *--               09/26/1994 -- Jay Parsons, removed size limits
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/10/1996 -- Modified to work with this class.
   *--                             Removed 'release' of array,
   *--                             changed to instantiate new array
   *--                             class, instead, which effectively
   *--                             does the same thing.
   *--               01/11/1997 -- changed to return array 
   *-- Usage.......: arrayex.txt2arr (<cFileName>)
   *-- Example.....: oArray = oArrayEx.txt2arr("CONFIG.DB")
   *-- Returns.....: array
   *-- Parameters..: cFileName  = Name of DOS Text file to read into 
   *--                            array
   *------------------------------------------------------------------
      
      local oArray, nSize, nHandle

      nSize = 0
      oArray = new array (nSize)
      
      *-- get file handle
      nHandle = 0
      nHandle = fopen( cFileName )
      IF nHandle > 0
         *-- store the file into the array
         local nLines
         nLines = 0
         DO WHILE .not. FEOF ( nHandle )
            nLines = nLines + 1
            IF nSize < nLines
               *-- make room in array 
               nSize = nSize + 100
               oArray.resize (nSize)
            ENDIF
            oArray [nLines] = FGETS (nHandle)
         ENDDO

         *-- close the file
         fClose( nHandle )
         oArray.resize (nLines)
      ENDIF
      
   RETURN oArray
   *-- EoM: Txt2Arr()

   FUNCTION Variance( aName, nRowCol, nWhich )
   *-------------------------------------------------------------------
   *-- Programmer..: Jay Parsons (CIS: 72662,1302)
   *-- Date........: 11/10/1996
   *-- Notes.......: Variance of non-blank numeric or date values in 
   *--               given row or column of a specified array.  If the 
   *--               first value is a date, considers only dates.  If 
   *--               first value is numeric or float, considers only 
   *--               numerics and floats.  Returns .F. if any value in 
   *--               row or column is any other type, if specified row 
   *--               or column does not exist or if there are no non
   *--               blank values. To adapt this to find the sample 
   *--               variance, substitute
   *--               "( nCount - 1 )" for the final "nCount" in the 
   *--               last line calculating xReturn.
   *-- Written for.: dBASE 5.0 for Windows
   *-- Rev. History: 10/01/1994 -- Original Release
   *--               11/10/1996 -- Modified to work with this class.
   *-- Calls.......: None
   *-- Usage.......: arrayex.Variance( <aName> [, <nRowCol> ;
   *--                                 [, <nWhich>]] )
   *-- Example.....: ? oMyArray.Variance( Myarray, 1, 1 )
   *-- Returns.....: A float value, the population variance.
   *-- Parameters..: aName   = the name of the array, not in quotes.
   *--               nRowCol = the number of the row or column to 
   *--                         average, default being 1
   *--               nWhich  = one of these values:
   *--                            1 = nRowCol specifies a row (default)
   *--                            2 = nRowCol specifies a column
   *-------------------------------------------------------------------
      
      local nW, nR, xReturn, cType, nAt, nTot, nCount, nTotSq, xNext, ;
            cType2
      
      nW = iif( nWhich = 2, 2, 1 )
      nR = iif( pcount() > 1, nRowCol, 1 )
      xReturn = .F.
      nDims = ( &aName..dimensions )
      if nDims = 1
         if nR # 1 .or. nW # 1
            RETURN xReturn
         else
            cType = type( "aName[ nR ]" )
         endif
      else
         if nW = 2
            cType = type( "aName[1,nR]" )
         else
            cType = type( "aName[nR,1]" )
         endif
      endif
      if .not. cType $ "DFN"
         RETURN xReturn
      endif
      store 0 to nTot, nCount, nTotSq
      for nAt = 1 to alen( aName, nW )
          do case
             case nDims = 1
                  xNext = aName[ nAt ]
             case nW = 1
                  xNext = aName[ nR, nAt ]
             otherwise
                  xNext = aName[ nAt, nR ]
          endcase
          cType2 = type( xNext )
          if ( .not. cType2 $ "DFN" ) .or. ;
             ( ( cType = "D" ) # ( cType2 = "D" ) )
             exit
          endif
          if isblank( xNext )
             loop
          endif
          xReturn = .T.
          if cType2 = "D"
             xNext = xNext - {01/01/01}
          endif
          nTot = nTot + xNext
          nTotSq = nTotSq + xNext * xNext
          nCount = nCount + 1
       endfor
       if xReturn # .F.
          xReturn = ( nTotSq - nTot * nTot / nCount ) / ;
                    nCount
       endif
       
    RETURN xReturn
    *-- EoM: Variance()

ENDCLASS
*-- End of Class: ArrayEx

#INCLUDE <duflp.h>

*-----------------------------------------------------------------------
*-- End of File: ARRAYEX.CC
*-----------------------------------------------------------------------