*-- File: FIELD.CC

CLASS Field
*----------------------------------------------------------------------
*-- Description:
*
*      This class contains methods designed to assist working with
*      fields.
*
*-- Programmers: 
*
*      Ken Mayer
*      Bowen Moursund (CIS: 72662,436)
*      Angus Scott-Fleming (CIS: 76711,517)
*      Oktay Amiry
*
*-- History: 
*
*      11/25/1996 -- original version of these routines as a class.
*
*-- Instantiation: 
*
*      set procedure to field.cc additive
*      _app.Field = new Field()
*
*-- Methods:
*      More details on each of these is contained in the commentary for
*      each method ...
*
*--    Release()
*        Releases this object from memory
*
*--    FldName()
*        Changes field name in .DBF using low-level file functions.
*
*--    IsMatch()
*        Checks for an index key match in named dbf or alias
*
*--    Mode()
*        Determines most common value for field
*
*-- Properties:
*
*      None
*----------------------------------------------------------------------

   FUNCTION Release
   *-------------------------------------------------------------------
   *-- Programmer..: Ken Mayer
   *-- Date........: 11/24/1996
   *-- Notes.......: Used to release this object from memory.
   *-- Written for.: Visual dBASE 5.5
   *-- Rev. History: 11/24/1996 -- Original Code
   *-- Calls.......: none
   *-- Usage.......: field.Release()
   *-- Example.....: _app.Field.Release()
   *--               _app.Field = NULL   && not required, but a good 
   *--                                   && idea
   *-- Returns.....: Null
   *-- Parameters..: None
   *-------------------------------------------------------------------

      release object this

   RETURN NULL
   *-- EoM: Release()


   FUNCTION FldName( cDbf, nField, cNewname )
   *-------------------------------------------------------------------
   *-- Programmer..: Bowen Moursund (CIS: 72662,436)
   *-- Date........: 11/10/1996
   *-- Notes.......: FldName() uses low level file functions to write
   *--               directly to a DBF header, changing the name of a
   *--               specified field.
   *-- Written for.: dBASE IV, version 1.5, 2.0
   *-- Rev. History: 07/12/1993 -- Original
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302                                     ,
   *--               11/10/1996 -- Modified to work with this class.
   *-- Calls.......: None
   *-- Usage.......: field.fldname( <cDbf>, <nField>, <cNewname> )
   *-- Example.....: cName = _app.Field.fldname( "MYDBF.DBF",5,"ADDRESS")
   *-- Returns.....: The new field name or a null string ("") on 
   *--               failure. 
   *-- Parameters..: cDbf    = The DBF name. Drive and path specs are 
   *--                         OK.
   *--                         The extension is required.
   *--               nField  = The field number.
   *--               cNewname= The new field name. Leading and trailing
   *--                         spaces will be trimmed, the name 
   *--                         truncated to 10 characters and converted
   *--                         to upper case.
   *-- WARNINGS....: NO checking for illegal characters in the new 
   *--               field name is made, so don't use any <g>. Since 
   *--               the DBF header is directly altered, a backup might
   *--               be desirable in the event of failure. If a field 
   *--               used in an index expression is changed, the index 
   *--               file should be rebuilt.
   *-------------------------------------------------------------------
      
      local cDbf, nField, cNew, nDbf, nNewloc, N, lSuccess, cRetstr
      lSuccess = .T.
      cNew = upper( left( ltrim( rtrim( cNewname ) ), 10 ) )
      use ( cDbf )
      nFields = fldcount()
      use
      lSuccess = ( nField <= nFields .and. nField > 0)
      if lSuccess
         nOffset = ( 32 * nField )
         nDbf = fopen( cDbf, "rw" )
         lSuccess = ( nDbf > 0 )
         if lSuccess
            nNewloc = fseek( nDbf, nOffset )
            lSuccess = ( nNewloc = nOffset )
            if lSuccess
               nBytes = fwrite( nDbf, replicate( chr(0), 11), 11 )
               nNewloc = fseek( nDbf, nOffset )
               lSuccess = ( nNewloc = nOffset )
               if lSuccess
                  nBytes = fwrite( nDbf, cNew )
               endif
            endif
         endif
         lSuccess = fclose( nDbf )
      endif
      cRetstr = iif( lSuccess, cNew, "" )
      
   RETURN cRetstr
   *-- EoM: FldName()

   FUNCTION IsMatch( cDbOrAlias, xValue, nRecno, cOrder )
   *-------------------------------------------------------------------
   *-- Programmers.: Bowen Moursund (CIS: 72662,436) and
   *--               Angus Scott-Fleming (CIS: 76711,517)
   *-- Date........: 11/10/1996
   *-- Notes.......: Checks for an index key match in the named DBF or
   *--               alias. Similar to version 2.0 KEYMATCH(), except 
   *--               that a match of the current record is ignored. 
   *--               The UDF may be used in 1 of 2 modes. If the 
   *--               optional parameter cOrder is passed to the UDF, 
   *--               then a copy of the named DBF will be opened in 
   *--               an unused work area, and the check for a key 
   *--               match made on that copy. Do not use this mode 
   *--               within a dBASE BEGIN/END TRANSACTION, as the
   *--               closing of the copy produces an error. If cOrder 
   *--               is NOT passed to the UDF, it's required that you 
   *--               USE the DBF twice, once for data entry and once 
   *--               for testing.
   *--               The second USE should include the keywords AGAIN
   *--               NOUPDATE:
   *--               use THEDBF order (cOrder) in select() alias XXX
   *--               use THEDBF order (cOrder) in select() alias YYY ;
   *--                                                  again noupdate
   *--               WARNING: if you use BEGIN/END TRANSACTION, you must 
   *--               open the testing database before using this 
   *--               function.
   *-- Written for.: dBASE IV, 1.5
   *-- Rev. History: 10/10/1993, Original Release.
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/10/1996 -- Modified to work with this class.
   *-- Calls.......: None
   *-- Usage.......: field.IsMatch(<cAlias>,<xValue>[,<nRecNo>]) OR
   *--               field.IsMatch(<cDBF>,<xValue>,<nRecNo>,<cOrder>)
   *-- Examples....: use with valid property of entryfield:
   *--                   valid {; _app.Field.isMatch(...) },;
   *--                   validrequired .t.
   *-- Returns.....: .T./.F.
   *-- Parameters..: REQUIRED
   *--               cDbOrAlias = Name or alias of DBF to check for 
   *--                            match
   *--               xValue     = Value (non-memo type) to check for 
   *--                            match
   *--               OPTIONAL
   *--               nRecNo     = The current record number.  Omit or 
   *--                            set to 0 if appending new records.
   *--               cOrder     = Optional parameter. Production MDX 
   *--                            Tag used to order the DBF if it's not
   *--                            already opened. Must allow SEEK on 
   *--                            field being checked.
   *-------------------------------------------------------------------
      
      local nPcount, lRetVal
      nPcount = pcount()
      do case
         case nPcount = 2  && DBF is already open and ordered; adding
              lRetVal = seek(xValue,cDbOrAlias)
         case nPcount = 3  && if DBF is already open and ordered
              lRetVal = seek(xValue,cDbOrAlias) .and. ;
                        (recno(cDbOrAlias) <> nRecno)
         case nPcount = 4  && need to open the DUPECHEK DBF
              local nNewArea
              nNewArea = select()
              use (cDbOrAlias) order tag cOrder) again in nNewArea ;
                  noupdate alias DUPECHEK
              lRetVal = seek(xValue,"DUPECHEK") .and. ;
                        (recno("DUPECHEK") <> nRecno)
              use in nNewArea
      endcase
      
   RETURN lRetVal
   *-- EoM: IsMatch()

   FUNCTION Mode( xField )
   *-------------------------------------------------------------------
   *-- Programmer..: Oktay Amiry
   *-- Date........: 11/18/1996
   *-- Notes.......: Used to determine the item which occurs most 
   *--               frequently in a list. 
   *--               Printed in TechNotes, December 1992.
   *-- Written for.: dBASE IV, 1.5
   *-- Rev. History: 12/01/1992 -- Original Release
   *--               Fall, 1994 -- revised for dBASE 5.0 for Windows.
   *--                             Jay Parsons, 72662,1302
   *--               11/18/1996 -- Modified to work with this class.
   *-- Calls.......: None
   *-- Usage.......: field.Mode(<xField>)
   *-- Example.....: ? _app.Field.Mode("SEX")
   *-- Returns.....: The item that is the most common among those in 
   *--               that field.
   *-- Parameters..: xField = an indexed field (it must be indexed)
   *-------------------------------------------------------------------
      
      local nCount, nMem, nOccur, nHigh, nName
      
      do case
         case tagcount() = 0
              RETURN "Specified file must be indexed"
         case reccount() <= 1
              RETURN "Invalid number of records for MODE()"
      endcase
      if upper(order()) # upper(xField)
         RETURN "Specified field must be indexed"
      endif
      
      go top
      nHigh = 1
      nCount = 0
      scan
         xCurrent = &xField.
         xSame = &xField.
         scan while xCurrent = xSame
              xCurrent = &xField.
              if xCurrent = xSame
                 nCount = nCount + 1
              endif
         endscan
         if nCount > nHigh
            nHigh = nCount
            xReturn = xSame
         else
            if nCount = nHigh
               xReturn = -1
            endif
         endif
         nCount = 0
      endscan
      
   RETURN iif(m->nHigh = 1, -1, m->xReturn)
   *-- EoM: Mode()

ENDCLASS
*-- End of Class: FIELD

*-----------------------------------------------------------------------
*-- End of File: FIELD.CC
*-----------------------------------------------------------------------